"""
Test suite for the elementary_rxn.py module

"""

import chemkin.reaction.elementary_rxn as er
import io
import sys


# tests for IrrevElemRxn() class
def test_IrrevElem_len_result():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    assert len(reac1) == 2


def test_IrrevElem_repr_result():
    old_stdout = sys.stdout
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    reac1 = er.IrreversibleElementaryRxn(ki=[10, 10], xi=[1.0, 2.0, 1.0], vi_p=[[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], vi_dp=[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])

    print(reac1)
    sys.stdout = old_stdout
    assert capturedOutput.getvalue() == "IrreversibleElementaryRxn(ki=[10, 10], xi=[1.0, 2.0, 1.0], vi_p=[[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], vi_dp=[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])\n"


def test_IrrevElem_progress_rate_result():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]],
                         [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    assert (reac1.progress_rate() == [40.0, 10.0]).all()


def test_IrrevElem_progress_rate_neg_ki():
    reac1 = er.IrreversibleElementaryRxn([-10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]]
                         ,[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.progress_rate()
    except ValueError as err: assert(type(err) == ValueError)


def test_IrrevElem_progress_rate_neg_xi():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [-1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.progress_rate()
    except ValueError as err:
        assert(type(err ) == ValueError)


def test_IrrevElem_progress_rate_dim_compat():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[ 2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.progress_rate()
    except ValueError as err:
        assert(type(err ) == ValueError)


def test_IrrevElem_reaction_rate_result():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0,
                         0.0, 2.0], [0.0, 1.0, 1.0]])

    assert (reac1.reaction_rate() == [-60., -70., 70.]).all()


def test_IrrevElem_reaction_rate_neg_ki():
    reac1 = er.IrreversibleElementaryRxn([-10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0, 0.0
                         , 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.reaction_rate()
    except ValueError as err:
        assert(type(err) == ValueError)


def test_IrrevElem_reaction_rate_neg_xi():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [-1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0, 0.0
                         , 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.reaction_rate()
    except ValueError as err:
        assert(type(err) == ValueError)


def test_IrrevElem_reaction_rate_dim_compat():
    reac1 = er.IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[2.0, 0.0], [2.0, 0.0, 2.0]],[[0.0, 0.0 , 2.0], [0.0, 1.0, 1.0]])
    try:
        reac1.reaction_rate()
    except ValueError as err:
        assert(type(err) == ValueError)


"""
Test suite for the base_rxn.py module

"""
import chemkin.reaction.base_rxn as rxn
import sys
import io

# tests for RxnBase() base class
def test_RxnBase_progress_rate_not_implemented():
    try:
        reac1 = rxn.RxnBase([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
        reac1.progress_rate()
    except NotImplementedError as err:
        assert(type(err) == NotImplementedError)


def test_RxnBase_reaction_rate_not_implemented():
    try:
        reac1 = rxn.RxnBase([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
        reac1.progress_rate()
    except NotImplementedError as err:
        assert(type(err) == NotImplementedError)


def test_RxnBase_len_result():
    reac1 = rxn.RxnBase([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    assert len(reac1) == 2


def test_RxnBase_repr_result():
    old_stdout = sys.stdout
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    reac1 = rxn.RxnBase([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
    print(reac1)
    sys.stdout = old_stdout
    assert capturedOutput.getvalue() == "RxnBase(ki=[10, 10], xi=[1.0, 2.0, 1.0], vi_p=[[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], vi_dp=[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])\n"

def test_RxnBase_reaction_rate():
    try:
        reac1 = rxn.RxnBase([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]])
        reac1.reaction_rate()
    except NotImplementedError as err:
        assert(type(err) == NotImplementedError)


from enum import Enum
import xml.etree.ElementTree as ET

import numpy as np

from chemkin.chemkin_errors import ChemKinError
from chemkin.reaction.reaction_coefficients import ArrheniusCoefficient, \
    ConstantCoefficient, ModifiedArrheniusCoefficient
from chemkin.thermodynamics.thermo import Thermo


class RxnType(Enum):
    Elementary = 1


class XmlParser():
    """Core method load() produces list of RxnData from XML file contents.

    Notes:
        ChemKinError raised when invalid values are encountered in the XML file.
    """

    def __init__ (self, path):
        """ Ensures path contains .xml file extension. """
        if path[-4:] != '.xml':
            path += '.xml'
        self.path = path

    def load (self):
        """ Parses XML file contents to create list of RxnData objects
        representing the reactions in the file.
        """
        tree = ET.parse(self.path)
        root = tree.getroot()
        species = []
        for species_i in root.find('phase'):
            new_species = species_i.text.strip().split()
            species.extend(new_species)

        results = []
        for rxn in root.find('reactionData').findall('reaction'):
            rxn_data = self.__extract_data_from_reaction_element(rxn)
            results.append(rxn_data)
        return species, results

    def __extract_data_from_reaction_element (self, rxn):
        """Returns RxnData object containing data from <reaction> XML element.
        Raises ChemKinError for invalid attribute/element values.
        """
        result = RxnData()
        result.rxn_id = rxn.get('id')

        # reversible
        reversible = rxn.get('reversible').lower().strip()
        if reversible in ['no', 'n', 'false', 'f']:
            result.reversible = False
        elif reversible in ['yes', 'y', 'true', 't']:
            result.reversible = True
        else:
            raise ChemKinError(
                  'XmlParser.load()',
                  'Invalid reversible attribute in reaction {}'.format(
                        result.rxn_id))

        # type
        rnx_type = rxn.get('type').lower().strip()
        if rnx_type == 'elementary':
            result.type = RxnType.Elementary
        else:
            raise ChemKinError(
                  'XmlParser.load()',
                  'Invalid type attribute in reaction {}'.format(
                        result.rxn_id))

        # rate_coeff
        rate_coeff = rxn.find('rateCoeff')
        if rate_coeff is None:
            raise ChemKinError('XmlParser.load()',
                               'No <rateCoeff> element found in one of the '
                               'reactions.')

        if rate_coeff.find('Arrhenius') is not None:
            arrhenius = rate_coeff.find('Arrhenius')
            A = float(arrhenius.find('A').text.strip())
            if A < 0:
                raise ChemKinError('XmlParser.load()',
                                   'A coeff < 0 in reaction with ' \
                                   'id = {}'.format(result.rxn_id))
            E = float(arrhenius.find('E').text.strip())
            result.rate_coeff = [A, E]
        elif rate_coeff.find('modifiedArrhenius') is not None:
            mod_arrhenius = rate_coeff.find('modifiedArrhenius')
            A = float(mod_arrhenius.find('A').text.strip())
            if A < 0:
                raise ChemKinError(
                      'A coeff < 0 in reaction with id = {}'.format(
                            result.rxn_id))
            b = float(mod_arrhenius.find('b').text.strip())
            E = float(mod_arrhenius.find('E').text.strip())
            result.rate_coeff = [A, b, E]
        elif rate_coeff.find('Constant') is not None:
            const = rate_coeff.find('Constant')
            result.rate_coeff = float(const.find('k').text.strip())
        else:
            raise ChemKinError('XmlParser.load()',
                               'No recognized child of <rateCoeff> found '
                               'from which to parse coefficients.')

        # reactants / products
        result.reactants = self.__map_conc_to_species(rxn.find('reactants'))
        result.products = self.__map_conc_to_species(rxn.find('products'))

        return result

    @staticmethod
    def __map_conc_to_species (tag):
        """ Creates dict mapping the species to concentration where tag is
        either a <reactants> or <products> XML tag.
        """
        result = {}
        for item in tag.text.strip().split():
            species, conc = item.strip().split(':')
            conc = int(conc)
            species = species.upper()
            result[species] = conc
        return result

    def parsed_data_list (self, Ti):
        species, rxn_data_list = self.load()
        n_species = len(species)

        species_idx_dict = {}  # build the dictionary of key = species_name,
        # value = species_index
        for i, s in enumerate(species):
            species_idx_dict[s] = i

        parsed_data_dic_list = []  # list of dicts, one for each 1 set of
        # rxns (1 xml file) under one temperature
        for T in Ti:
            sys_vi_p = []  # list of reactant Stoichiometric coefficients in
            # each rxn
            sys_vi_dp = []  # list of product Stoichiometric coefficients in
            # each rxn
            ki = []  # list of reation rate coefficients in each rxn
            is_reversible = None  # indicator of the system of reactions
            # being irreversible/reversible

            for rxn_data in rxn_data_list:  # 1 rxn per rxn_data
                if rxn_data.type != RxnType.Elementary:
                    raise ChemKinError('XmlParser.parsed_data_list(Ti)',
                                       'Non-elementary reactions cannot be '
                                       'parsed now.')

                if is_reversible == None:
                    is_reversible = rxn_data.reversible  # set the indicator
                    # of the system of reactions to be irreversible/reversible

                if rxn_data.reversible != is_reversible:  # the system of
                    # reactions in the XML file must be all
                    # irreversible/reversible
                    raise ChemKinError('XmlParser.parsed_data_list(Ti)',
                                       'The system of reactions in the XML '
                                       'file {} are inconsistent in '
                                       'reversibility.'.format(
                                           self.path))

                rxn_id = rxn_data.rxn_id  # save id

                rxn_vi_p = np.zeros((
                                    n_species,))  # save the Stoichiometric
                # coefficients of the reactants in this rxn
                for s, vi in rxn_data.reactants.items():
                    idx = species_idx_dict[s]  # get index of the specii
                    rxn_vi_p[idx] = vi
                sys_vi_p.append(list(rxn_vi_p))

                rxn_vi_dp = np.zeros((
                                     n_species,))  # save the Stoichiometric
                # coefficients of the products in this rxn
                for s, vi in rxn_data.products.items():
                    idx = species_idx_dict[s]  # get index of the specii
                    rxn_vi_dp[idx] = vi
                sys_vi_dp.append(list(rxn_vi_dp))

                coef_params = rxn_data.rate_coeff
                if isinstance(coef_params, list):
                    if len(coef_params) == 3:  # modified arrhenius coef
                        A = coef_params[0]
                        b = coef_params[1]
                        E = coef_params[2]
                        ki.append(
                            ModifiedArrheniusCoefficient(A, b, E, T).get_coef())
                    else:  # arrhenius coef
                        A = coef_params[0]
                        E = coef_params[1]
                        ki.append(ArrheniusCoefficient(A, E, T).get_coef())
                else:  # const coef
                    ki.append(ConstantCoefficient(coef_params).get_coef())

                parsed_data_dic = {}
                parsed_data_dic['species'] = species
                parsed_data_dic['ki'] = ki
                parsed_data_dic['sys_vi_p'] = sys_vi_p
                parsed_data_dic['sys_vi_dp'] = sys_vi_dp
                parsed_data_dic['is_reversible'] = is_reversible
                parsed_data_dic['T'] = T

            if is_reversible == True:
                try:
                    b_ki = Thermo(species, T, ki, sys_vi_p,
                                  sys_vi_dp).get_backward_coefs()
                    parsed_data_dic['b_ki'] = b_ki
                except ChemKinError as err:
                    print(str(err))
                    parsed_data_dic['b_ki'] = 'Not Defined'

            parsed_data_dic_list.append(parsed_data_dic)

        return parsed_data_dic_list


class RxnData():
    """ Container for individual reaction data.

    Attributes
    ----------
    rxn_id : str
        id attribute of <reaction> element.
    reversible : bool
        True if reaction is reversible; False if irreversible.
    reactants : Dict[str, int]
        Mapping of species to stoichiometric coefficients for reactants.
        Example: {'H2':1, 'O':1}
    products : Dict[str, int]
        Mapping of species to stoichiometric coefficients for products.
    rate_coeff : List[float] or float
        Reaction rate coefficients contained depend on the type of rate
        coefficients in the XML file, dictated by the child of the <rateCoeff>
        element:
            Arrhenius: [A, E]
            modifiedArrhenius: [A, b, E]
            Constant: k
    type : RxnType
        Enum value for reaction type.
    """

    def __init__ (self, rxn_id=None, reversible=None, reactants=None,
                  products=None, rate_coeff=None, type=None):
        self.rxn_id = rxn_id
        self.reversible = reversible
        self.reactants = reactants
        self.products = products
        self.rate_coeff = rate_coeff
        self.type = type

    def equation (self):
        """ Returns equation representation of reactants and products.

        Species are listed in alphabetical order on both reactant and product
        side.
        """
        reactant_side = self.__build_equation_side(self.reactants)
        product_side = self.__build_equation_side(self.products)
        return '{0} [=] {1}'.format(reactant_side, product_side)

    @staticmethod
    def __build_equation_side (species_dict):
        result = ''
        species_count = -1
        for species, conc in sorted(species_dict.items()):
            species_count += 1

            # Handle first species differently than rest, then proceed to
            # next species.
            if species_count == 0:
                if abs(conc) == 1:
                    conc_part = ''
                else:
                    conc_part = conc
                result += '{0}{1}'.format(conc_part, species)
                continue

            # Non-first species.
            if conc < 0:
                operator = ' - '
            else:
                operator = ' + '

            if abs(conc) == 1:
                conc_part = ''
            else:
                conc_part = abs(conc)

            result += '{0}{1}{2}'.format(operator, conc_part, species)

        return result



import os

from chemkin.chemkin_errors import *
# Import all module attributes into chemkin namespace.
from chemkin.preprocessing.parse_xml import *
from chemkin.reaction.base_rxn import *
from chemkin.reaction.elementary_rxn import *
from chemkin.reaction.non_elementary_rxn import *
from chemkin.reaction.reaction_coefficients import *
from chemkin.thermodynamics.thermo import *

# Define path to XML files based on relative position to this module.
_PATH_XML_FILES = os.path.join(os.path.dirname(__file__), 'xml-files/')


def pckg_xml_path (file_name):
    """Creates full path to XML file located in xml-files directory."""
    if file_name[-4:] != '.xml':
        file_name += '.xml'
    return _PATH_XML_FILES + file_name



import numpy as np
from chemkin.reaction.base_rxn import RxnBase


class ElementaryRxn(RxnBase):
    """Class of elementary reactions
    Subclass of RxnBase
    """
    pass


class ReversibleElementaryRxn(ElementaryRxn):
    """Class of Reversible elementary reactions
    Subclass of ElemRxn
    Default form of
            v_11' A + v_21' B <-> v_31" C
            v_32' C <-> v_12' A + v_22' B

    ATTRIBUTES:
    ========
    self.ki: float or a list of floats, optional, default value = [10.0, 10.0]
        Forward reaction rate coefficient
        Initialized with constructor argument ki

    self.b_ki: float or a list of floats, optional, default value = [20.0, 20.0]
        Backward reaction rate coefficient
        Initialized with constructor argument b_ki

    self.xi: a list of floats, optional, default value = [1.0, 1.0, 1.0]
        Concentrations of molecular species
        Initialized with constructor argument xi

    self.vi_p: a list of floats, optional, default value = [[1.0, 1.0, 0.0],
    [0.0, 0.0, 1.0]]
        Stoichiometric coefficients of the reactants
        Initialized with constructor argument vi_p

    self.vi_dp: a list of floats, optional, default value = [[0.0, 0.0, 1.0],
    [1.0, 1.0, 0.0]]
        Stoichiometric coefficients of the products
        Initialized with constructor argument vi_dp


    NOTES
    =====
    PRE:
        - ki, b_ki, xi, vi_p, and vi_dp have list or np.array type
        - xi is ordered in the form [[A], [B], [C]]
        - vi_p is ordered in the form [[v_11', v_21', v_31'], [v_12', v_22', v_32']]
        - vi_dp is ordered in the form [[v_11", v_21", v_31"], [v_12", v_22", v_32"]]
        - four or fewer inputs
    METHODS:
    ========
    __len__(): Returns the number of reactions
    __repr__(): Prints the class name with its attributes
    progress_rate(): Calculates and returns the total progress rate (forward progress rate - backward progress rate)
    reaction_rate(): Calculates and returns the reaction rate
    """
    def __init__ (self, ki=[10.0, 10.0], b_ki=[20.0, 20.0], xi=[1.0, 1.0, 1.0],
                  vi_p=[[1.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                  vi_dp=[[0.0, 0.0, 1.0], [1.0, 1.0, 0.0]]):
        self.ki = ki
        self.b_ki = b_ki
        self.xi = xi
        self.vi_p = vi_p
        self.vi_dp = vi_dp
        self.wi = None
        self.rates = None

    def __len__ (self):
        """Returns the number of reactions"""
        return len(self.vi_p)

    def __repr__ (self):
        return 'ReversibleElementaryRxn(ki={}, b_ki={}, xi={}, vi_p={}, vi_dp={})'.format(self.ki,
                                                                            self.b_ki,
                                                                            self.xi,
                                                                            self.vi_p,
                                                                            self.vi_dp)

    def progress_rate (self):
        """
        Returns the progress rate w for a system of reversible elementary
        reaction
        
        RETURNS
        ========
        self.wi: a numpy array of floats,
           Has the form self.w unless self.ki <= 0 or self.b_ki <= 0 or self.xi < 0
           in which cases a ValueError exception is raised
        
        NOTES
        =====
        POST:
             - self.ki, self.xi, self.vi_p, and self.vi_dp are not changed by
             this function
             - raises a ValueError exception if any(self.ki <= 0)
             - raises a ValueError exception if any(self.b_ki <= 0)
             - raises a ValueError exception if any(self.xi < 0)
             - returns a numpy array of floats of progress rate self.wi
        
        EXAMPLES
        =========

        MAKE EXAMPLE DOC TEST !!!!!! 


        """

        # check value conditions
        if (any(i <= 0 for i in self.ki)):  # check forward reaction coefficients
            raise ValueError("forward reaction rate coefficients ki must be positive.")
        elif (any(i <= 0 for i in self.b_ki)):  # check backward reaction coefficients
            raise ValueError("backward reaction rate coefficients b_ki must be positive.")
        elif (any(i < 0 for i in self.xi)):  # check concentration array
            raise ValueError("concentrations xi cannot be negative.")
        else:
            # Convert inputs into numpy column vectors/matrices
            xi = np.array(self.xi).reshape(-1, 1)
            vi_p = np.matrix(self.vi_p).T
            vi_dp = np.matrix(self.vi_dp).T
            ki = np.array(self.ki).reshape(-1, 1)
            b_ki = np.array(self.b_ki).reshape(-1, 1)

            f_prod = np.prod(np.power(xi, vi_p), axis=0)  # calculate the product of xi^(vi_p) for each reaction
            b_prod = np.prod(np.power(xi, vi_dp), axis=0)  # calculate the product of xi^(vi_p) for each reaction
            
            f_wi = np.squeeze(ki * np.array(f_prod).reshape(-1, 1)) # forward progress rate
            b_wi = np.squeeze(b_ki * np.array(b_prod).reshape(-1, 1)) # backward progress rate
            self.wi = f_wi - b_wi  # set total progress rate 

            return self.wi

    def reaction_rate (self):
        """Returns the progress rate w for a system of reversible elementary reaction

        RETURNS
        ========
        self.rates: a numpy array of floats,
           Has the form self.rates unless self.ki <= 0 or self.b_ki <= 0 or self.xi < 0
           in which cases a ValueError exception is raised
    
        NOTES
        =====
        POST:
             - self.ki, self.b_ki, self.xi, self.vi_p, and self.vi_dp are not changed by this function
             - raises a ValueError exception if any(self.ki <= 0)
             - raises a ValueError exception if any(self.b_ki <= 0)
             - raises a ValueError exception if any(self.xi < 0)
             - returns a numpy array of floats of reaction rates, self.rates
        
        EXAMPLES
        =========

        MAKE EXAMPLE DOC TEST !!!!!! 

        """
        # check value conditions
        if any(i <= 0 for i in self.ki):  # check forward reaction coefficients
            raise ValueError("forward reaction rates ki must be positive.")
        elif any(i <= 0 for i in self.b_ki):  # check backward reaction coefficients
            raise ValueError("backward reaction rates ki must be positive.")
        elif any(i < 0 for i in self.xi):  # check concentration array
            raise ValueError("concentrations xi cannot be negative.")
        else:
            # Convert inputs into numpy column vectors/matrices
            vi_p = np.matrix(self.vi_p).T
            vi_dp = np.matrix(self.vi_dp).T
            vi = vi_dp - vi_p  # calculate overall stoicheometric coefficients

            if self.wi == None:
                self.progress_rate()
            w = self.wi  # get progress rate

            self.rates = np.squeeze(np.array(np.dot(vi, w)))  # calculate reaction rate

            return self.rates


class IrreversibleElementaryRxn(ElementaryRxn):
    """Class of irreversible elementary reactions
    Subclass of ElementaryRxn
    Default form of
            v_11' A + v_21' B -> v_31" C
            v_32' C -> v_12' A + v_22' B

    ATTRIBUTES:
    ========
    self.ki: float or a list of floats, optional, default value = [10.0, 10.0]
        Reaction rate coefficient
        Initialized with constructor argument ki

    self.xi: a list of floats, optional, default value = [1.0, 1.0, 1.0]
        Concentrations of molecular species
        Initialized with constructor argument xi

    self.vi_p: a list of floats, optional, default value = [[1.0, 1.0, 0.0],
    [0.0, 0.0, 1.0]]
        Stoichiometric coefficients of the reactants
        Initialized with constructor argument vi_p

    self.vi_dp: a list of floats, optional, default value = [[0.0, 0.0, 1.0],
    [1.0, 1.0, 0.0]]
        Stoichiometric coefficients of the products
        Initialized with constructor argument vi_dp

    NOTES
    =====
    PRE:
        - ki, xi, vi_p, and vi_dp have list or np.array type
        - xi is ordered in the form [[A], [B], [C]]
        - vi_p is ordered in the form [[v_11', v_21', v_31'], [v_12', v_22',
        v_32']]
        - vi_dp is ordered in the form [[v_11", v_21", v_31"], [v_12", v_22",
        v_32"]]
        - four or fewer inputs
    METHODS:
    ========
    __len__(): Returns the number of reactions
    __repr__(): Prints the class name with its attributes
    progress_rate(): Calculates and returns the progress rate
    reaction_rate(): Calculates and returns the reaction rate
    """

    def __init__(self, ki=[10.0, 10.0], xi=[1.0, 1.0, 1.0],
                 vi_p=[[1.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                 vi_dp=[[0.0, 0.0, 1.0], [1.0, 1.0, 0.0]]):
        self.ki = ki
        self.xi = xi
        self.vi_p = vi_p
        self.vi_dp = vi_dp
        self.wi = None
        self.rates = None

    # def __len__(self):
    #     """Returns the number of species in the reaction"""
    #     return len(self.xi)

    def __len__(self):
        """Returns the number of reactions"""
        return len(self.vi_p)

    def __repr__(self):
        return 'IrreversibleElementaryRxn(ki={}, xi={}, vi_p={}, vi_dp={})'.format(self.ki,
                                                                      self.xi,
                                                                      self.vi_p,
                                                                      self.vi_dp)

    def progress_rate(self):
        """
        Returns the progress rate w for a system of irreversible elementary
        reaction

        RETURNS
        ========
        self.wi: a numpy array of floats,
           Has the form self.w unless self.ki <= 0 or self.xi < 0
           in which cases a ValueError exception is raised

        NOTES
        =====
        POST:
             - self.ki, self.xi, self.vi_p, and self.vi_dp are not changed by
             this function
             - raises a ValueError exception if any(self.ki <= 0)
             - raises a ValueError exception if any(self.xi < 0)
             - returns a numpy array of floats of progress rate self.wi

        EXAMPLES
        =========
        >>> IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0,0.0, 2.0]],[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]]).progress_rate()
        array([ 40.,  10.])
        """

        # check value conditions
        if (any(i <= 0 for i in self.ki)):  # check reaction coefficients
            raise ValueError("reaction rate coefficients ki must be positive.")
        elif (any(i < 0 for i in self.xi)):  # check concentration array
            raise ValueError("concentrations xi cannot be negative.")
        else:
            # Convert inputs into numpy column vectors/matrices
            xi = np.array(self.xi).reshape(-1, 1)
            vi_p = np.matrix(self.vi_p).T
            ki = np.array(self.ki).reshape(-1, 1)

            prodi = np.prod(np.power(xi, vi_p),
                            axis=0)  # calculate the product of xi^(vi_p) for
            #  each reaction
            self.wi = np.squeeze(
                ki * np.array(prodi).reshape(-1, 1))  # multiply ki by prodi

            return self.wi

    def reaction_rate(self):
        """Returns the progress rate w for a system of irreversible
        elementary reaction

        RETURNS
        ========
        self.rates: a numpy array of floats,
           Has the form self.rates unless self.ki <= 0 or self.xi < 0
           in which cases a ValueError exception is raised

        NOTES
        =====
        POST:
             - self.ki, self.xi, self.vi_p, and self.vi_dp are not changed by
             this function
             - raises a ValueError exception if any(self.ki <= 0)
             - raises a ValueError exception if any(self.xi < 0)
             - returns a numpy array of floats of reaction rates, self.rates

        EXAMPLES
        =========
        >>> IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0,0.0, 2.0]],[[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]]).reaction_rate()
        array([-60., -70.,  70.])
        """
        # check value conditions
        if any(i <= 0 for i in self.ki):  # check reaction coefficients
            raise ValueError("reaction rates ki must be positive.")
        elif any(i < 0 for i in self.xi):  # check concentration array
            raise ValueError("concentrations xi cannot be negative.")
        else:
            # Convert inputs into numpy column vectors/matrices
            vi_p = np.matrix(self.vi_p).T
            vi_dp = np.matrix(self.vi_dp).T

            w = self.progress_rate()  # calculate progress rate
            vi = vi_dp - vi_p  # calculate overall stoicheometric coefficients

            self.rates = np.squeeze(
                np.array(np.dot(vi, w)))  # calculate reaction rate

            return self.rates

#a = IrreversibleElementaryRxn([10, 10], [1.0, 2.0, 1.0], [[1.0, 2.0, 0.0], [2.0, 0.0, 2.0]], [[0.0, 0.0, 2.0], [0.0, 1.0, 1.0]]).progress_rate()
#print(a)



class RxnBase():
    """Base class of reactions

    ATTRIBUTES:
    ========
    self.ki: float or a list of floats, required
        Reaction rate coefficient
        Initialized with constructor argument ki

    self.xi: a list of floats, required,
        Concentrations of molecular species
        Initialized with constructor argument xi

    self.vi_p: a list of floats, required,
        Stoichiometric coefficients of the reactants
        Initialized with constructor argument vi_p

    self.vi_dp: a list of floats, required,
        Stoichiometric coefficients of the products
        Initialized with constructor argument vi_dp

    METHODS:
    ========
    __len__(): Returns the number of reactions
    __repr__(): Prints the class name with its attributes

    progress_rate(): Calculates and returns the progress rate
        This should be implemented in its subclass.

    reaction_rate(): Calculates and returns the reaction rate
        This should be implemented in its subclass.
    """

    def __init__ (self, ki, xi, vi_p, vi_dp):
        self.ki = ki
        self.xi = xi
        self.vi_p = vi_p
        self.vi_dp = vi_dp
        self.wi = None
        self.rates = None

    # def __len__(self):
    #     """Returns the number of species in the reaction"""
    #     return len(self.xi)

    def __len__ (self):
        """Returns the number of reactions"""
        return len(self.vi_p)

    def __repr__ (self):
        return 'RxnBase(ki={}, xi={}, vi_p={}, vi_dp={})'.format(self.ki, self.xi,
                                                             self.vi_p,
                                                             self.vi_dp)

    def progress_rate (self):
        """ Calculates and returns the progress rate

        Method is not implemented in this base class, but should be
        implemented in its subclasses.
        """
        raise NotImplementedError('Subclass must implement this method')

    def reaction_rate (self):
        """ Calculates and returns the reaction rate

        Method is not implemented in this base class, but should be
        implemented in its subclasses.
        """
        raise NotImplementedError('Subclass must implement this method')